import logging

from django.core.management.base import BaseCommand
from django.db.models.fields.related import ManyToManyField
from django.db.models.fields.related import OneToOneField

from django_neuralyzer.base import BaseNeuralyzer
from django_neuralyzer.utils import get_app_submodules


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)]
    )


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = ""

    def handle(self, *args, **options):
        # ensure neuralyzers has been loaded
        list(get_app_submodules("neuralyzers"))
        # get all subclasses
        anon_classes = all_subclasses(BaseNeuralyzer)
        errors = []

        for klass in anon_classes:
            print(f"Neuralyzer: {klass}")
            model = klass.Meta.model
            neuralyzer = klass()
            anon_fields = neuralyzer._get_class_attributes()
            noop_fields = neuralyzer._excluded_attributes
            model_fields = model._meta.fields
            model_fields_names = [field.name for field in model_fields]
            for field in model_fields:
                if isinstance(field, (ManyToManyField, OneToOneField)):
                    continue

                if field.name not in anon_fields and field.name not in noop_fields:
                    errors.append(
                        f"Neuralyzer {klass.__name__} is missing field {field.name} for model {model.__name__}"
                    )
            for field in anon_fields:
                if field not in model_fields_names:
                    errors.append(
                        f"Neuralyzer {klass.__name__} has extra field {field} for model {model.__name__}"
                    )

        if errors:
            logger.error("Following models have not been fully handled: \n" + "\n".join(errors))
        else:
            print("All models neuralyzed include all fields!")



from django.db.models import Q
from django.test import TestCase

from django_neuralyzer.base import NEURALYZER_NOOP
from django_neuralyzer.base import BaseNeuralyzer as BaseBaseNeuralyzer
from django_neuralyzer.base import lazy_attribute

from . import models


class BaseNeuralyzer(BaseBaseNeuralyzer):
    class Meta:
        pass

    def __init__(self):
        """
        The reason we set declarations manually in ``__init__`` is because we want to
        patch objects individually for unit testing purposes, without using the
        ``run()`` method. In general cases, declarations will be set by ``run()``
        """
        self._declarations = self.get_declarations()


class BaseTestCase(TestCase):
    def test_get_manager(self):
        class Neuralyzer(BaseNeuralyzer):
            class Meta:
                model = models.Person
                manager = models.PersonAnotherQuerySet

        neuralyzer = Neuralyzer()
        self.assertEqual(neuralyzer.get_manager(), models.PersonAnotherQuerySet)

    def test_get_queryset(self):
        sample_obj = models.person_factory()

        class Neuralyzer(BaseNeuralyzer):
            class Meta:
                model = models.Person

        neuralyzer = Neuralyzer()
        result = neuralyzer.get_queryset()
        self.assertSequenceEqual(result, [sample_obj])

    def test_patch_object(self):
        fake_first_name = "foo"

        class Neuralyzer(BaseNeuralyzer):
            first_name = fake_first_name
            last_name = fake_first_name
            raw_data = "{1: 2}"

        obj = models.person_factory(last_name="")  # empty data should be kept empty

        neuralyzer = Neuralyzer()
        neuralyzer.patch_object(obj)
        self.assertEqual(obj.first_name, "foo")
        self.assertEqual(obj.last_name, "")
        self.assertEqual(obj.raw_data, "{1: 2}")

    def test_inheritance(self):
        class ParentNeuralyzer(BaseNeuralyzer):
            first_name = "parent"
            last_name = "parent"

        class ChildNeuralyzer(ParentNeuralyzer):
            last_name = "child"

        obj = models.person_factory()

        neuralyzer = ChildNeuralyzer()
        neuralyzer.patch_object(obj)
        self.assertEqual(obj.first_name, "parent")
        self.assertEqual(obj.last_name, "child")

    def test_run(self):
        class Neuralyzer(BaseNeuralyzer):
            class Meta:
                model = models.Person

            first_name = "xyz"

        obj = models.person_factory()

        neuralyzer = Neuralyzer()
        neuralyzer.run()

        obj.refresh_from_db()
        self.assertEqual(obj.first_name, "xyz")

    def test_run_without_fields(self):
        class Neuralyzer(BaseNeuralyzer):
            def clean(self, obj):
                obj.first_name = "xyz"
                obj.save()

            class Meta:
                model = models.Person

        obj = models.person_factory()

        neuralyzer = Neuralyzer()
        neuralyzer.run()

        obj.refresh_from_db()
        self.assertEqual(obj.first_name, "xyz")

    def test_run_filters_dict(self):
        class Neuralyzer(BaseNeuralyzer):
            first_name = "xyz"

            class Meta:
                model = models.Person

        obja = models.person_factory()
        objb = models.person_factory()

        neuralyzer = Neuralyzer()
        neuralyzer.run(filters={"pk": 1})

        obja.refresh_from_db()
        objb.refresh_from_db()
        self.assertEqual(obja.first_name, "xyz")
        self.assertEqual(objb.first_name, "A")

    def test_run_filters_Q(self):
        class Neuralyzer(BaseNeuralyzer):
            first_name = "xyz"

            class Meta:
                model = models.Person

        obja = models.person_factory()
        objb = models.person_factory()

        neuralyzer = Neuralyzer()
        neuralyzer.run(filters=~Q(pk=1))

        obja.refresh_from_db()
        objb.refresh_from_db()
        self.assertEqual(obja.first_name, "A")
        self.assertEqual(objb.first_name, "xyz")

    def test_lazy_attribute(self):
        fake_first_name = lazy_attribute(lambda o: o.last_name)

        class Neuralyzer(BaseNeuralyzer):
            first_name = fake_first_name

        obj = models.person_factory()

        neuralyzer = Neuralyzer()
        neuralyzer.patch_object(obj)
        self.assertEqual(obj.first_name, obj.last_name)

    def test_lazy_attribute_decorator(self):
        class Neuralyzer(BaseNeuralyzer):
            @lazy_attribute
            def first_name(self):
                return "xyz"

        obj = models.person_factory()

        neuralyzer = Neuralyzer()
        neuralyzer.patch_object(obj)
        self.assertEqual(obj.first_name, "xyz")

    def test_raw_attributes(self):
        class Neuralyzer(BaseNeuralyzer):
            raw_data = "{}"

        obj = models.person_factory()

        neuralyzer = Neuralyzer()
        neuralyzer.patch_object(obj)
        self.assertEqual(obj.raw_data, "{}")

    def test_clean(self):
        class Neuralyzer(BaseNeuralyzer):
            line1 = ""
            line2 = ""
            line3 = ""

            def clean(self, obj):
                obj.line1 = "foo"
                obj.line2 = "bar"

        obj = models.person_factory()

        neuralyzer = Neuralyzer()
        neuralyzer.patch_object(obj)
        self.assertEqual(obj.line1, "foo")
        self.assertEqual(obj.line2, "bar")
        self.assertEqual(obj.line3, "")

    def test_get_declarations(self):
        # Ensure the order is preserved
        class Neuralyzer(BaseNeuralyzer):
            a = lazy_attribute(lambda o: 4)
            c = lazy_attribute(lambda o: 6)
            b = lazy_attribute(lambda o: 5)

            def get_queryset(self):
                return []

        neuralyzer = Neuralyzer()
        self.assertEqual(list(neuralyzer.get_declarations().keys()), ["a", "c", "b"])

    def test_excluded_attributes(self):
        class Neuralyzer(BaseNeuralyzer):
            first_name = NEURALYZER_NOOP

        obj = models.person_factory()

        neuralyzer = Neuralyzer()
        self.assertIn("first_name", neuralyzer._excluded_attributes)
        self.assertNotIn("first_name", neuralyzer._get_class_attributes())



from collections import OrderedDict
import inspect
from logging import getLogger

from django.db.models import Q

from .utils import import_from_path

logger = getLogger(__name__)

NEURALYZER_NOOP = "__NOOP__"


class OrderedDeclaration(object):
    """Any classes inheriting from this will have an unique global counter
    associated with it. This counter is used to determine the order in
    which fields were declarated

    Idea taken from: https://stackoverflow.com/a/4460034/639465
    Also inspired by https://github.com/FactoryBoy/factory_boy
    """

    global_counter = 0

    def __init__(self):
        self._order = self.__class__.global_counter
        self.__class__.global_counter += 1


class LazyAttribute(OrderedDeclaration):
    def __init__(self, lazy_fn):
        super(LazyAttribute, self).__init__()
        self.lazy_fn = lazy_fn

    def __call__(self, *args, **kwargs):
        return self.lazy_fn(*args, **kwargs)


def lazy_attribute(lazy_fn):
    """Returns LazyAttribute objects, that basically marks functions that
    should take `obj` as first parameter. This is useful when you need
    to take in consideration other values of `obj`

    Example:

    >>> full_name = lazy_attribute(o: o.first_name + o.last_name)

    """
    return LazyAttribute(lazy_fn)


class BaseNeuralyzer(object):
    def run(self, filters=None, select_chunk_size=None, **bulk_update_kwargs):
        self._declarations = self.get_declarations()

        queryset = self.get_queryset(filters=filters)
        update_fields = list(self._declarations.keys())

        # info used in log messages
        model_name = self.Meta.model.__name__

        logger.info("Updating {}...".format(model_name))

        manager = self.get_manager()
        objs = []
        for obj in queryset:
            self.patch_object(obj)
            objs.append(obj)

        if update_fields:
            manager.bulk_update(
                objs,
                update_fields,
                **dict(**bulk_update_kwargs),
            )
        else:
            logger.info("Skiping bulk update for {}... No fields to update".format(model_name))

        # Cascade to one to one relation
        if hasattr(self.Meta, "onetoone"):
            for relation, class_import in self.Meta.onetoone.items():
                neuralyzer = import_from_path(class_import)
                for obj in objs:
                    related_model = getattr(obj, relation)
                    if related_model:
                        neuralyzer().run(filters={"pk": related_model.pk})

    def get_manager(self):
        meta = self.Meta
        return getattr(meta, "manager", meta.model.objects)

    def get_queryset(self, filters=None):
        """Override this if you want to delimit the objects that should be
        affected by anonymization

        Args:
          filters: Q object or dict
        """
        qs = self.get_manager().all()
        if isinstance(filters, dict):
            qs = qs.filter(**filters)
        elif isinstance(filters, Q):
            qs = qs.filter(filters)
        return qs

    def patch_object(self, obj):
        """Update object attributes with fake data provided by replacers"""
        fields = [field for field in self._declarations]

        for field in fields:
            replacer = self._declarations[field]
            if isinstance(replacer, LazyAttribute):
                # Pass in obj for LazyAttributes
                new_value = replacer(obj)
            elif callable(replacer):
                new_value = replacer()
            else:
                new_value = replacer

            setattr(obj, field, new_value)

        self.clean(obj)

    def clean(self, obj):
        """Use this function if you need to update additional data that may
        rely on multiple fields, or if you need to update multiple fields
        at once
        """
        pass

    def get_declarations(self):
        """Returns ordered declarations. Any non-ordered declarations, for
        example any types that does not inherit from OrderedDeclaration
        will come first, as they are considered "raw" values and should
        not be affected by the order of other non-ordered declarations
        """

        def _sort_declaration(declaration):
            name, value = declaration
            if isinstance(value, OrderedDeclaration):
                return value._order
            else:
                # Any non-ordered declarations come first
                return -1

        declarations = self._get_class_attributes().items()
        sorted_declarations = sorted(declarations, key=_sort_declaration)

        return OrderedDict(sorted_declarations)

    @property
    def _excluded_attributes(self):
        reserved_names = list(BaseNeuralyzer.__dict__.keys()) + ["Meta", "_declarations"]
        return [
            name
            for name, value in inspect.getmembers_static(self)
            if not name.startswith("__")
            and name not in reserved_names
            and value in [NEURALYZER_NOOP]
        ]

    def _get_class_attributes(self):
        """Return list of class attributes, which also includes methods and
        subclasses, ignoring any magic methods and reserved attributes
        as well as defined noop attributes
        """
        reserved_names = list(BaseNeuralyzer.__dict__.keys()) + ["Meta", "_declarations"]

        return {
            name: getattr(self, name)
            for name, value in inspect.getmembers_static(self)
            if not name.startswith("__")
            and name not in reserved_names
            and value not in [NEURALYZER_NOOP]
        }

    class Meta:
        noop = []
        onetoone = {}
